@@ -13,39 +13,77 @@ Module to watch the execution of shell scripts. Both streams (`stdout` and
     watch.run(['rsync', '-av', '/home', '/backup'])
 """
 
+from __future__ import annotations
+
 import abc
 import logging
 import os
 import pwd
 import queue
 import shlex
+import shutil
 import socket
 import subprocess
-import shutil
 import sys
 import textwrap
 import threading
-import typing
 import time
+import typing
 import uuid
-
 from logging.handlers import BufferingHandler
+from typing import (IO, Any, Dict, List, Literal, Optional, Sequence, Tuple,
+                    TypedDict, Union, cast)
 
-from . import termcolor, send_nsca
-from .config_reader import ConfigReader
+from typing_extensions import Unpack
+
+from . import capturing, icinga, termcolor
+from .config_reader import ClassInterface, ConfigReader, Spec
 from .send_email import send_email
 
-
 HOSTNAME = socket.gethostname()
 USERNAME = pwd.getpwuid(os.getuid()).pw_name
 
 
+Status = Literal[0, 1, 2, 3]
+
+
+class MinimalMessageParams(TypedDict, total=False):
+
+    custom_message: str
+    """Custom message"""
+
+    prefix: str
+    """ Prefix of the report message."""
+
+    body: str
+    """ A longer report text."""
+
+    preformance_data: Dict[str, Any]
+    """ A dictionary like
+          `{'perf_1': 1, 'perf_2': 'test'}`"""
+
+
+class MessageParams(MinimalMessageParams, total=False):
+
+    status: Status
+    """ 0 (OK), 1 (WARNING), 2 (CRITICAL), 3 (UNKOWN): see
+          Nagios / Icinga monitoring status / state."""
+
+    service_name: str
+    """The name of the service."""
+
+    log_records: str
+    """Log records separated by new lines"""
+
+    processes: List['Process']
+
+
 class BaseClass:
 
-    def _obj_to_str(self, attributes=[]):
+    def _obj_to_str(self, attributes: List[str] = []) -> str:
         if not attributes:
             attributes = dir(self)
-        output = []
+        output: List[str] = []
         for attribute in attributes:
             if not attribute.startswith('_') and \
                not callable(getattr(self, attribute)):
@@ -60,11 +98,11 @@ class BaseClass:
 class CommandWatcherError(Exception):
     """Exception raised by this module."""
 
-    def __init__(self, msg, **data):
+    def __init__(self, msg: str, **data: Unpack[MessageParams]):
         reporter.report(
             status=2,
             custom_message='{}: {}'.format(self.__class__.__name__, msg),
-            **data,
+            **data,  # type: ignore
         )
 
 
@@ -115,12 +153,14 @@ class LoggingHandler(BufferingHandler):
     """Store of all logging records in the memory. Print all records on emit.
     """
 
-    def __init__(self, master_logger=None):
+    _master_logger:  Optional[logging.Logger]
+
+    def __init__(self, master_logger: Optional[logging.Logger] = None):
         BufferingHandler.__init__(self, capacity=1000000)
         self._master_logger = master_logger
 
     @staticmethod
-    def _print(record):
+    def _print(record: logging.LogRecord):
         """
         :param logging.LogRecord record: A record object.
         """
@@ -193,7 +233,7 @@ class LoggingHandler(BufferingHandler):
 
     @property
     def stdout(self):
-        messages = []
+        messages: List[str] = []
         for record in self.buffer:
             if record.levelname == 'STDOUT':
                 messages.append(record.msg)
@@ -201,7 +241,7 @@ class LoggingHandler(BufferingHandler):
 
     @property
     def stderr(self):
-        messages = []
+        messages: List[str] = []
         for record in self.buffer:
             if record.levelname == 'STDERR':
                 messages.append(record.msg)
@@ -210,30 +250,39 @@ class LoggingHandler(BufferingHandler):
     @property
     def all_records(self):
         """All log messages joined by line breaks."""
-        messages = []
+        messages: List[str] = []
         for record in self.buffer:
             messages.append(self.format(record))
         return '\n'.join(messages)
 
 
-def _log_stdout(self, message, *args, **kws):
+class ExtendedLogger(logging.Logger):
+
+    def stdout(self, line: object, *args: Any, **kws: Any) -> None: ...
+    def stderr(self, line: object, *args: Any, **kws: Any) -> None: ...
+
+
+def _log_stdout(self: ExtendedLogger, message: object, *args: Any, **kws: Any):
     # Yes, logger takes its '*args' as 'args'.
     self._log(STDOUT, message, args, **kws)
 
 
-logging.Logger.stdout = _log_stdout
+extendedLogger: ExtendedLogger = cast(ExtendedLogger, logging.Logger)
 
 
-def _log_stderr(self, message, *args, **kws):
+extendedLogger.stdout = _log_stdout  # type: ignore
+
+
+def _log_stderr(self: ExtendedLogger, message: object, *args: Any, **kws: Any):
     # Yes, logger takes its '*args' as 'args'.
     self._log(STDERR, message, args, **kws)
 
 
-logging.Logger.stderr = _log_stderr
+logging.Logger.stderr = _log_stderr  # type: ignore
 
 
-def setup_logging(master_logger: logging.Logger = None) -> \
-        typing.Tuple[logging.Logger, LoggingHandler]:
+def setup_logging(master_logger: Optional[logging.Logger] = None) -> \
+        typing.Tuple[ExtendedLogger, LoggingHandler]:
     """Setup a fresh logger for each watch action.
 
     :param master_logger: Forward all log messages to a master logger."""
@@ -249,7 +298,7 @@ def setup_logging(master_logger: logging.Logger = None) -> \
     # the root logger is created with level WARNING.
     logger.setLevel(1)
     logger.addHandler(handler)
-    return (logger, handler)
+    return (cast(ExtendedLogger, logger), handler)
 
 
 # Reporting ###################################################################
@@ -259,18 +308,11 @@ class Message(BaseClass):
     """
     This message class bundles all available message data into an object. The
     different reporters can choose which data they use.
-
-    :param int status: 0 (OK), 1 (WARNING), 2 (CRITICAL), 3 (UNKOWN): see
-        Nagios / Icinga monitoring status / state.
-    :param str service_name: The name of the service.
-    :param str custom_message: Custom message
-    :param str prefix: Prefix of the report message.
-    :param str body: A longer report text.
-    :param dict performance_data: A dictionary like
-        `{'perf_1': 1, 'perf_2': 'test'}`.
-    :param str log_records: Log records separated by new lines
     """
-    def __init__(self, **data):
+
+    _data: MessageParams
+
+    def __init__(self, **data: Unpack[MessageParams]):
         self._data = data
 
     def __str__(self):
@@ -285,7 +327,7 @@ class Message(BaseClass):
     @property
     def status_text(self) -> str:
         """The status as a text word like `OK`."""
-        return send_nsca.States[self.status]
+        return icinga.States[self.status]
 
     @property
     def service_name(self) -> str:
@@ -298,9 +340,11 @@ class Message(BaseClass):
         :rtype: str
         """
         performance_data = self._data.get('performance_data')
-        if performance_data:
-            pairs = []
-            for key, value in self._data.get('performance_data').items():
+        if performance_data and isinstance(performance_data, dict):
+            pairs: List[str] = []
+            key: str
+            value: Any
+            for key, value in performance_data.items():
                 pairs.append('{!s}={!s}'.format(key, value))
             return ' '.join(pairs)
         return ''
@@ -315,7 +359,7 @@ class Message(BaseClass):
 
     @property
     def message(self) -> str:
-        output = []
+        output: List[str] = []
         if self.prefix:
             output.append(self.prefix)
 
@@ -328,7 +372,7 @@ class Message(BaseClass):
     @property
     def message_monitoring(self) -> str:
         """message + performance_data"""
-        output = []
+        output: List[str] = []
         output.append(self.message)
         if self.performance_data:
             output.append('|')
@@ -338,7 +382,7 @@ class Message(BaseClass):
     @property
     def body(self) -> str:
         """Text body for the e-mail message."""
-        output = []
+        output: List[str] = []
         output.append('Host: {}'.format(HOSTNAME))
         output.append('User: {}'.format(USERNAME))
         output.append('Service name: {}'.format(self.service_name))
@@ -346,7 +390,7 @@ class Message(BaseClass):
         if self.performance_data:
             output.append('Performance data: {}'.format(self.performance_data))
 
-        body = self._data.get('body', '')
+        body: str = self._data.get('body', '')
         if body:
             output.append('')
             output.append(body)
@@ -361,8 +405,8 @@ class Message(BaseClass):
         return '\n'.join(output)
 
     @property
-    def processes(self) -> str:
-        output = []
+    def processes(self) -> Optional[str]:
+        output: List[str] = []
         processes = self._data.get('processes')
         if processes:
             for process in processes:
@@ -379,8 +423,7 @@ class BaseChannel(BaseClass, metaclass=abc.ABCMeta):
     """Base class for all reporters"""
 
     @abc.abstractmethod
-    def report(self, status: int = 0, service_name: str = 'command_watcher',
-               **data):
+    def report(self, message: Message) -> None:
         raise NotImplementedError('A reporter class must have a `report` '
                                   'method.')
 
@@ -425,37 +468,33 @@ class EmailChannel(BaseChannel):
         )
 
 
-class NscaChannel(BaseChannel):
-    """Wrapper around `send_nsca` to send NSCA messages. Set up the NSCA
-    client."""
+class IcingaChannel(BaseChannel):
+    url: str
+    user: str
+    password: str
+    service_name: str
 
-    def __init__(self, remote_host: str, password: str, encryption_method: int,
-                 port: int, service_name: str):
-        self.remote_host = remote_host
+    def __init__(self, url: str, user: str, password: str,
+                 service_name: str):
+        self.url = url
+        self.user = user
         self.password = password
-        self.encryption_method = encryption_method
-        self.port = port
         self.service_name = service_name
 
     def __str__(self):
         # No password!
-        return self._obj_to_str(['remote_host', 'encryption_method', 'port',
-                                 'service_name'])
+        return self._obj_to_str(['url', 'user', 'service_name'])
 
     def report(self, message: Message):
-        """Send a NSCA message to a remote NSCA server.
-
-        :param message: A message object.
-        """
-        send_nsca.send_nsca(
+        icinga.send_passive_check(
+            url=self.url,
+            user=self.user,
+            password=self.password,
             status=message.status,
             host_name=HOSTNAME,
             service_name=message.service_name,
-            text_output=message.message_monitoring,
-            remote_host=self.remote_host,
-            password=str(self.password),
-            encryption_method=self.encryption_method,
-            port=self.port,
+            text_output=message.message,
+            performance_data=message.performance_data
         )
 
 
@@ -502,13 +541,15 @@ class BeepChannel(BaseChannel):
 class Reporter:
     """Collect all channels."""
 
+    channels: List[BaseChannel]
+
     def __init__(self):
         self.channels = []
 
-    def add_channel(self, channel):
+    def add_channel(self, channel: BaseChannel):
         self.channels.append(channel)
 
-    def report(self, **data):
+    def report(self, **data: Unpack[MessageParams]):
         message = Message(**data)
         for channel in self.channels:
             channel.report(message)
@@ -529,7 +570,7 @@ CONF_DEFAULTS = {
 }
 
 
-CONFIG_READER_SPEC = {
+CONFIG_READER_SPEC: Spec = {
     'email': {
         'from_addr': {
             'description': 'The email address of the sender.',
@@ -576,6 +617,21 @@ CONFIG_READER_SPEC = {
             'default': 5667,
         },
     },
+    'icinga': {
+        'url': {
+            'description': 'The HTTP URL. /v1/actions/process-check-result '
+                           'is appended.',
+            'not_empty': True,
+        },
+        'user': {
+            'description': 'The user for the HTTP authentification.',
+            'not_empty': True,
+        },
+        'password': {
+            'description': 'The password for the HTTP authentification.',
+            'not_empty': True,
+        },
+    },
     'beep': {
         'activated': {
             'description': 'Activate the beep channel to report auditive '
@@ -588,6 +644,22 @@ CONFIG_READER_SPEC = {
 
 # Main code ###################################################################
 
+Args = Union[str, List[str], Tuple[str]]
+
+
+class ProcessArgs(TypedDict, total=False):
+    shell: bool
+    """If true, the command will be executed through the
+        shell.
+    """
+
+    cwd: str
+    """Sets the current directory before the child is
+        executed."""
+
+    env: Dict[str, Any]
+    """Defines the environment variables for the new process."""
+
 
 class Process:
     """Run a process.
@@ -597,30 +669,30 @@ class Process:
 
     :param args: List, tuple or string. A sequence of
         process arguments, like `subprocess.Popen(args)`.
-    :param master_logger:
-    :param bool shell: If true, the command will be executed through the
-        shell.
-    :param str cwd: Sets the current directory before the child is
-        executed.
-    :param dict env: Defines the environment variables for the new process.
     """
-    def __init__(self, args: typing.Union[str, list, tuple],
-                 master_logger: logging.Logger = None, **kwargs):
+    args: Args
+    """Process arguments in various types."""
+
+    _queue: 'queue.Queue[Optional[Tuple[bytes, capturing.Stream]]]'
+
+    log: ExtendedLogger
+    """A ready to go and configured logger."""
+
+    log_handler: LoggingHandler
+
+    subprocess: subprocess.Popen[Any]
+
+    def __init__(self, args: Args,
+                 master_logger: Optional[ExtendedLogger] = None,
+                 **kwargs: Unpack[ProcessArgs]):
         # self.args: typing.Union[str, list, tuple] = args
         self.args = args
-        """Process arguments in various types."""
 
         self._queue = queue.Queue()
-        """An instance of :py:class:`queue.Queue`."""
 
         log, log_handler = setup_logging(master_logger=master_logger)
-        # self.log: logging.Logger = log
         self.log = log
-        """A ready to go and configured logger. An instance of
-        :py:class:`logging.Logger`."""
-        # self.log_handler: LoggingHandler = log_handler
         self.log_handler = log_handler
-        """An instance of :py:class:`LoggingHandler`."""
 
         self.log.info('Run command: {}'.format(' '.join(self.args_normalized)))
         timer = Timer()
@@ -633,7 +705,6 @@ class Process:
             # bufsize=1,
             **kwargs
         )
-        """subprocess"""
 
         self._start_thread(self.subprocess.stdout, 'stdout')
         self._start_thread(self.subprocess.stderr, 'stderr')
@@ -652,7 +723,7 @@ class Process:
         self.log.info('Execution time: {}'.format(timer.result()))
 
     @property
-    def args_normalized(self) -> list:
+    def args_normalized(self) -> Sequence[str]:
         """Normalized `args`, always a list"""
         if isinstance(self.args, str):
             return shlex.split(self.args)
@@ -679,10 +750,9 @@ class Process:
         """The count of lines of the current `stderr`."""
         return len(self.stderr.splitlines())
 
-    def _stdout_stderr_reader(self, pipe, stream):
+    def _stdout_stderr_reader(self, pipe: IO[bytes], stream: capturing.Stream):
         """
         :param object pipe: `process.stdout` or `process.stdout`
-        :param str stream: `stdout` or `stderr`
         """
         try:
             with pipe:
@@ -691,10 +761,9 @@ class Process:
         finally:
             self._queue.put(None)
 
-    def _start_thread(self, pipe, stream):
+    def _start_thread(self, pipe: Optional[IO[str]], stream: capturing.Stream):
         """
         :param object pipe: `process.stdout` or `process.stdout`
-        :param str stream: `stdout` or `stderr`
         """
         threading.Thread(
             target=self._stdout_stderr_reader,
@@ -714,27 +783,47 @@ class Watch:
     :param config_reader: A custom configuration reader. Specify this
       parameter to not use the build in configuration reader.
     """
-    def __init__(self, config_file: str = None,
+
+    _hostname: str
+    """The hostname of machine the watcher running on."""
+
+    _service_name: str
+    """A name of the watched service."""
+
+    log: ExtendedLogger
+    """A ready to go and configured logger."""
+
+    _log_handler: LoggingHandler
+
+    processes: List[Process]
+    """A list of completed processes
+    :py:class:`Process`. Everytime you use the method
+    `run()` the process object is appened in the list."""
+
+    _conf: Optional[ClassInterface]
+
+    _raise_exceptions: bool
+    """Raise exceptions"""
+
+    _timer: Timer
+
+    def __init__(self, config_file: Optional[str] = None,
                  service_name: str = 'command_watcher',
                  raise_exceptions: bool = True,
-                 config_reader: ConfigReader = None,
-                 report_channels: typing.List[BaseChannel] = None):
+                 config_reader: Optional[ConfigReader] = None,
+                 report_channels: Optional[List[BaseChannel]] = None):
         self._hostname = HOSTNAME
-        """The hostname of machine the watcher running on."""
 
         self._service_name = service_name
-        """A name of the watched service."""
 
         log, log_handler = setup_logging()
+
         self.log = log
-        """A ready to go and configured logger. An instance of
-        :py:class:`logging.Logger`."""
         self.log.info('Hostname: {}'.format(self._hostname))
+
         self._log_handler = log_handler
-        """An instance of :py:class:`LoggingHandler`."""
 
         self._conf = None
-        """An instance of :py:class:`jflib.config_reader.ClassInterface`."""
 
         if not config_reader and config_file:
             config_reader = ConfigReader(
@@ -743,6 +832,9 @@ class Watch:
                 dictionary=CONF_DEFAULTS,
             )
 
+        if not config_reader:
+            raise Exception('No config_reader object')
+
         self._conf = config_reader.get_class_interface()
 
         if report_channels is None:
@@ -762,16 +854,15 @@ class Watch:
                 pass
 
             try:
-                config_reader.check_section('nsca')
-                nsca_reporter = NscaChannel(
-                    remote_host=self._conf.nsca.remote_host,
-                    password=self._conf.nsca.password,
-                    encryption_method=self._conf.nsca.encryption_method,
-                    port=self._conf.nsca.port,
+                config_reader.check_section('icinga')
+                icinga_reporter = IcingaChannel(
+                    url=self._conf.icinga.url,
+                    user=self._conf.icinga.user,
+                    password=self._conf.icinga.password,
                     service_name=self._service_name,
                 )
-                reporter.add_channel(nsca_reporter)
-                self.log.debug(nsca_reporter)
+                reporter.add_channel(icinga_reporter)
+                self.log.debug(icinga_reporter)
             except (ValueError, KeyError):
                 pass
 
@@ -784,12 +875,8 @@ class Watch:
             reporter.channels = []
 
         self.processes = []
-        """A list of completed processes
-        :py:class:`Process`. Everytime you use the method
-        `run()` the process object is appened in the list."""
 
         self._raise_exceptions = raise_exceptions
-        """Raise exceptions"""
 
         self._timer = Timer()
 
@@ -803,9 +890,11 @@ class Watch:
         """Alias / shortcut for `self._log_handler.stderr`."""
         return self._log_handler.stderr
 
-    def run(self, args: typing.Union[str, list, tuple],
-            log: bool = True, ignore_exceptions: typing.List[int] = [],
-            **kwargs) -> Process:
+    def run(self,
+            args: Args,
+            log: bool = True,
+            ignore_exceptions: List[int] = [],
+            **kwargs: Unpack[ProcessArgs]) -> Process:
         """
         Run a process.
 
@@ -833,16 +922,9 @@ class Watch:
             )
         return process
 
-    def report(self, status: int, **data) -> Message:
+    def report(self, status: Status,
+               **data: Unpack[MinimalMessageParams]) -> Message:
         """Report a message using the preconfigured channels.
-
-        :param int status: 0 (OK), 1 (WARNING), 2 (CRITICAL), 3 (UNKOWN): see
-          Nagios / Icinga monitoring status / state.
-        :param str custom_message: Custom message
-        :param str prefix: Prefix of the report message.
-        :param str body: A longer report text.
-        :param dict performance_data: A dictionary like
-          `{'perf_1': 1, 'perf_2': 'test'}`.
         """
         message = reporter.report(
             status=status,
@@ -854,24 +936,16 @@ class Watch:
         self.log.debug(message)
         return message
 
-    def final_report(self, **data) -> Message:
+    def final_report(self, **data: Unpack[MessageParams]) -> Message:
         """The same as the `report` method. Adds `execution_time` to the
         `performance_data`.
-
-        :param int status: 0 (OK), 1 (WARNING), 2 (CRITICAL), 3 (UNKOWN): see
-          Nagios / Icinga monitoring status / state.
-        :param str custom_message: Custom message
-        :param str prefix: Prefix of the report message.
-        :param str body: A longer report text.
-        :param dict performance_data: A dictionary like
-          `{'perf_1': 1, 'perf_2': 'test'}`.
         """
         timer_result = self._timer.result()
         self.log.info(
             'Overall execution time: {}'.format(timer_result)
         )
         status = data.get('status', 0)
-        data_dict = dict(data)
+        data_dict: Dict[str, Any] = dict(data)
         if 'performance_data' not in data_dict:
             data_dict['performance_data'] = {}
         data_dict['performance_data']['execution_time'] = timer_result
